(function (GCN) {

	'use strict';

	/**
	 * Maps constructcategories that were fetched via the Rest API 
	 * into a sorted nested array of constructs.
	 *
	 * @param {object<string, object>} constructs
	 * @return {object<string, object>}
	 */
	function mapConstructCategories(constructs) {
		var constructKeyword;
		var categoryMap = { categories: {}, categorySortorder: [] };
		var constructCategoryArray = [];

		for (constructKeyword in constructs) {
			if (constructs.hasOwnProperty(constructKeyword)) {
				var construct = constructs[constructKeyword];

				var constructCategoryName = construct.category;
				var categorySortorder = construct.categorySortorder;

				// Use a custom name for constructs that have not been assigned to a category
				if (!constructCategoryName) {
					constructCategoryName = "GCN_UNCATEGORIZED";
					categorySortorder = -1;
				}

				// Initialize the inner array of constructs
				if (!categoryMap.categories[constructCategoryName]) {
					var newCategory = {}; 
					newCategory.constructs = {};
					newCategory.sortorder = categorySortorder;
					newCategory.name = constructCategoryName;
					categoryMap.categories[constructCategoryName] = newCategory;
					constructCategoryArray.push(newCategory);
				}

				// Add the construct to the category
				categoryMap.categories[constructCategoryName].constructs[constructKeyword] = construct;
			}
		}

		// Sort the categories by the sortorder
		constructCategoryArray.sort(function (a, b) {
			return a.sortorder - b.sortorder;
		});

		// Add the sorted category names to the sortorder field
		var k;
		for (k in constructCategoryArray) {
			if (constructCategoryArray.hasOwnProperty(k)) {
				var category = constructCategoryArray[k];
				if (typeof category.sortorder !== 'undefined' && category.sortorder !== -1) {
					categoryMap.categorySortorder.push(category.name);
				}
			}
		}
		
		return categoryMap;
	}
	
	/**
	 * Maps constructs that were fetched via the Rest API into a flat hash map.
	 *
	 * @param {object<string, object>} constructs
	 * @return {object<string, object>}
	 */
	function mapConstructs(constructs) {
	
		var map = {};
		
		if (!constructs) {
			return map;
		}
		
		var constructId;
		for (constructId in constructs) {
			if (constructs.hasOwnProperty(constructId)) {
				var construct = constructs[constructId];
				map[construct.keyword] = construct;
			}
		}
		return map;
	}
	

	/**
	 * Node object.
	 *
	 * @name NodeAPI
	 * @class
	 * @augments Chainback
	 */
	var NodeAPI = GCN.defineChainback({
		/** @lends NodeAPI */

		__chainbacktype__: 'NodeAPI',
		_extends: GCN.ContentObjectAPI,
		_type: 'folder',

		_data: {
			folderId: null
		},

		/**
		 * @private
		 * @type {object<string, number} Constructs for this node are cached
		 *                               here so that we only need to fetch
		 *                               this once.
		 */
		_constructs: null,

		/**
		 * @private
		 * @type {object<string, object} Constructs categories for this node.
		 *                               Cached here so that we only need to
		 *                               fetch this once.
		 */
		_constructCategories: null,

		/**
		 * Retrieves a list of constructs and constructs categories that are
		 * assigned to this node and passes it as the only argument into the
		 * the `success()' callback.
		 *
		 * @param {function(Array.<object>)=} success Callback to receive an
		 *                                            array of constructs.
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 * @return Returns the constructs / categories
		 * @throws INVALID_ARGUMENTS
		 */
		constructs: function (success, error) {
			if (!success) {
				GCN.error('INVALID_ARGUMENTS', 'the `constructs()\' method ' +
					'requires at least a success callback to be given');
			}

			if (this._constructs) {
				success(this._constructs);
			} else {
				var that = this;
				this._authAjax({
					url   : GCN.settings.BACKEND_PATH +
					        '/rest/construct/list.json',
					type  : 'GET',
					error : function (xhr, status, msg) {
						GCN.handleHttpError(xhr, msg, error);
					},
					success: function (response) {
						if (GCN.getResponseCode(response) === 'OK') {
							that._constructs = mapConstructs(response.constructs);
							//console.dir(that._constructs);
							success(that._constructs);
						} else {
							GCN.handleResponseError(response, error);
						}
					}
				});
			}
		},

		/**
		 * Removes this node object.
		 *
		 * @param {function=} success Callback.
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 * @param {function} success callback
		 */
		remove: function (success, error) {

		},

		/**
		 * @FIXME: Is it really possible to save changes to a node?  If not,
		 *         then we should not surface this method.
		 */
		save: function () {},

		/**
		 * Retreives the top-level folders of this node's root folder.
		 *
		 * @param {function(FolderAPI)=} success
		 * @param {function(GCNError):boolean=} error Custom error handler.
		 */
		'!folders': function (success, error) {
			return this.folder(null, error).folders(success, error);
		},

		/**
		 * Helper method that will load the constructs of this node.
		 *
		 * @private
		 * @this {NodeAPI}
		 * @param {function(Array.<object>)} success callback
		 * @param {function(GCNError):boolean=} error callback
		 */
		constructCategories: function (success, error) {
			if (this._constructCategories) {
				success(this._constructCategories);
			} else {
				var that = this;
				this._continueWith(function (child) {
					that._data.id = child._data.nodeId;

					that.constructs(function (constructs) {
						that._constructCategories = mapConstructCategories(constructs);
						success(that._constructCategories);
					}, error);
				}, error);
			
				/*	
				this._authAjax({
					url   : GCN.settings.BACKEND_PATH +
					        '/rest/construct/load.json',
					type  : 'GET',
					error : function (xhr, status, msg) {
						GCN.handleHttpError(xhr, msg, error);
					},
					success: function (response) {
						if (GCN.getResponseCode(response) === 'OK') {
							that._constructCategories =
								response.constructCategories;
							success(response.constructCategories);
						} else {
							GCN.handleResponseError(response, error);
						}
					}
				});
				*/
			}
		}

	});

	GCN.node = GCN.exposeAPI(NodeAPI);
	GCN.NodeAPI = NodeAPI;

}(GCN));
